/********************************************************************/
/*                                                                  */
/*  drw_rtl.c     Generic graphic drawing functions.                */
/*  Copyright (C) 1989 - 2013  Thomas Mertes                        */
/*                                                                  */
/*  This file is part of the Seed7 Runtime Library.                 */
/*                                                                  */
/*  The Seed7 Runtime Library is free software; you can             */
/*  redistribute it and/or modify it under the terms of the GNU     */
/*  Lesser General Public License as published by the Free Software */
/*  Foundation; either version 2.1 of the License, or (at your      */
/*  option) any later version.                                      */
/*                                                                  */
/*  The Seed7 Runtime Library is distributed in the hope that it    */
/*  will be useful, but WITHOUT ANY WARRANTY; without even the      */
/*  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR */
/*  PURPOSE.  See the GNU Lesser General Public License for more    */
/*  details.                                                        */
/*                                                                  */
/*  You should have received a copy of the GNU Lesser General       */
/*  Public License along with this program; if not, write to the    */
/*  Free Software Foundation, Inc., 51 Franklin Street,             */
/*  Fifth Floor, Boston, MA  02110-1301, USA.                       */
/*                                                                  */
/*  Module: Seed7 Runtime Library                                   */
/*  File: seed7/src/drw_rtl.c                                       */
/*  Changes: 2007, 2013  Thomas Mertes                              */
/*  Content: Generic graphic drawing functions.                     */
/*                                                                  */
/********************************************************************/

#include "version.h"

#include "stdlib.h"
#include "stdio.h"
#include "string.h"

#include "common.h"
#include "data_rtl.h"
#include "rtl_err.h"
#include "drw_drv.h"

#undef EXTERN
#define EXTERN
#include "drw_rtl.h"



void drwCpy (wintype *win_to, wintype win_from)

  { /* drwCpy */
#ifdef TRACE_DRW
    printf("drwCpy(%lu, %ld)\n", win_to, win_from);
#endif
    if (win_from != NULL) {
      win_from->usage_count++;
    } /* if */
    if (*win_to != NULL) {
      (*win_to)->usage_count--;
      if ((*win_to)->usage_count == 0) {
        drwFree(*win_to);
      } /* if */
    } /* if */
    *win_to = win_from;
  } /* drwCpy */



wintype drwCreate (wintype win_from)

  { /* drwCreate */
    if (win_from != NULL) {
      win_from->usage_count++;
    } /* if */
    return win_from;
  } /* drwCreate */



/**
 *  Generic Create function to be used via function pointers.
 *  Function pointers in C programs generated by the Seed7 compiler
 *  may point to this function. This assures correct behaviour even
 *  when sizeof(rtlGenerictype) != sizeof(wintype).
 */
rtlGenerictype drwCreateGeneric (const rtlGenerictype from_value)

  { /* drwCreateGeneric */
    return (rtlGenerictype) (memsizetype)
           drwCreate((wintype) (memsizetype) from_value);
  } /* drwCreateGeneric */



void drwDestr (wintype old_win)

  { /* drwDestr */
    if (old_win != NULL) {
      old_win->usage_count--;
      if (old_win->usage_count == 0) {
        drwFree(old_win);
      } /* if */
    } /* if */
  } /* drwDestr */



wintype drwRtlImage (const const_rtlArraytype image)

  {
    const_rtlObjecttype *curr_line;
    const_rtlObjecttype *curr_column;
    rtlArraytype arr_line;
    int32type *pixel_elem;
    memsizetype height;
    memsizetype width;
    memsizetype line;
    memsizetype column;
    int32type *image_data;
    wintype result;

  /* drwRtlImage */
    height = arraySize(image);
    /* printf("drwRtlImage: height=%d\n", height); */
    if (height == 0) {
      raise_error(RANGE_ERROR);
      result = NULL;
    } else {
      curr_line = &image->arr[0];
      arr_line = curr_line->value.arrayvalue;
      width = arraySize(arr_line);
      /* printf("drwRtlImage: width=%d\n", width); */
      if (width == 0) {
        raise_error(RANGE_ERROR);
        result = NULL;
      } else {
        if (height > MAX_MEMSIZETYPE / sizeof(int32type) / (memsizetype) width ||
            (image_data = (int32type *) malloc((memsizetype) height * (memsizetype) width *
                                               sizeof(int32type))) == NULL) {
          raise_error(MEMORY_ERROR);
          result = NULL;
        } else {
          pixel_elem = image_data;
          for (line = height; line > 0; line--, curr_line++) {
            arr_line = curr_line->value.arrayvalue;
            curr_column = &arr_line->arr[0];
            for (column = width; column > 0; column--, curr_column++) {
              *pixel_elem = (int32type) curr_column->value.intvalue;
              pixel_elem++;
            } /* for */
          } /* for */
          result = drwImage(image_data, width, height);
          free(image_data);
        } /* if */
      } /* if */
    } /* if */
    return result;
  } /* drwRtlImage */
